module data-view where

open import Data.Maybe hiding (maybe)
open import Data.Unit
open import Data.Unit1
open import Data.Product
open import Data.Nat
open import Data.Sum
open import Data.List hiding ([_];fromMaybe)
open import Data.Vec  hiding ([_];fromList;toList)
open import Data.Vec1
open import Data.Bool
open import Relation.Binary.PropositionalEquality
open import Relation.Binary.PropositionalEquality1

open import local-lib
infixr 50 _⇒_

data Kind : Set where
  ⋆    : Kind
  _⇒_ : Kind → Kind → Kind

⟦_⟧ : Kind → Set1
⟦ ⋆ ⟧ = Set0
⟦ a ⇒ b ⟧ = ⟦ a ⟧ → ⟦ b ⟧

data Const : Kind → Set where
  Nat   : Const ⋆
  Unit  : Const ⋆
  Sum   : Const (⋆ ⇒ ⋆ ⇒ ⋆)
  Prod  : Const (⋆ ⇒ ⋆ ⇒ ⋆)

interp-c : ∀ {k} → Const k → ⟦ k ⟧
interp-c Unit  = ⊤
interp-c Nat   = ℕ
interp-c Sum   = _⊎_
interp-c Prod  = _×_

Ctx : Set
Ctx = List Kind

mutual
  data TyVar : Ctx → Kind → Set where
    VZ   : ∀ {G k} → TyVar (k ∷ G) k
    VS   : ∀ {G k' k} → TyVar G k → TyVar (k' ∷ G) k 

  data Typ : Ctx → Kind → Set1 where 
    Var  : ∀ {G k} → TyVar G k → Typ G k 
    Lam  : ∀ {G k1 k2} → Typ (k1 ∷ G) k2 
      → Typ G (k1 ⇒ k2)
    App  : ∀ {G k1 k2} → Typ G (k1 ⇒ k2) 
      → Typ G k1
      → Typ G k2
    Con  : ∀ {G k} → Const k → Typ G k
    Data : ∀ {G} → DT G → Typ G ⋆

  data Env : List Kind → Set1 where
    []  : Env [] 
    _∷_ : ∀ {k G} → ⟦ k ⟧ → Env G 
       → Env (k ∷ G)

  data DT (G : Ctx) : Set1 where
    mkDT  :  (t     : Typ G ⋆)
          →  (s     : Env G → Set)
          →  (to    : ({e : Env G} → interp t e → s e))
          →  (from  : ({e : Env G} → s e → interp t e))
          →  DT G

  sLookup : ∀ {k G} → TyVar G k → Env G → ⟦ k ⟧
  sLookup VZ     (v ∷ G) = v
  sLookup (VS x) (v ∷ G) = sLookup x G

  interp : ∀ {k G} → Typ G k → Env G →  ⟦ k ⟧
  interp (Var x) e       = sLookup x e
  interp (Lam t) e       = \ y → interp t (y ∷ e)
  interp (App t1 t2) e   = (interp t1 e) (interp t2 e)
  interp (Con c) e       = interp-c c
  interp (Data dt) e     = DT-s dt e

  DT-t : {G : Ctx} → DT G → Typ G ⋆
  DT-t  (mkDT t _ _ _) = t 
  DT-s : {G : Ctx} → DT G → Env G → Set
  DT-s    (mkDT _ s _ _) = s
  DT-to : {G : Ctx} → (dt : DT G) 
    → ({e : Env G} → interp (DT-t dt) e → (DT-s dt) e)
  DT-to   (mkDT _ _ to _) = to
  DT-from : {G : Ctx} → (dt : DT G) 
    → ({e : Env G} → (DT-s dt) e → interp (DT-t dt) e)
  DT-from (mkDT _ _ _ from) = from

Ty : Kind → Set1
Ty = Typ []

⌊_⌋ : {k : Kind} → Ty k → ⟦ k ⟧
⌊ t ⌋ = interp t []

data Orange : Set where
   toOrange : ℕ → Orange

fromOrange : Orange → ℕ
fromOrange (toOrange x) = x

orange : {G : Ctx} → Typ G ⋆
orange = Data (mkDT 
           (Con Nat)              -- t 
           (\ e → Orange)         -- s 
           (\ {e} → toOrange)     -- to
           (\ {e} → fromOrange))  -- from

maybeDef : {G : Ctx} → Typ G ⋆ → Typ G ⋆
maybeDef t = (App (App (Con Sum) (Con Unit)) t)

toMaybe  : {G : Ctx} {e : Env G} 
         → (t : Typ G ⋆)
         → (interp (maybeDef t) e) → Maybe (interp t e) 
toMaybe t (inj₁ _)  = nothing
toMaybe t (inj₂ x)  = just x

fromMaybe  : {G : Ctx} {e : Env G} → (t : Typ G ⋆)
           → Maybe (interp t e) 
           → (interp (maybeDef t) e)
fromMaybe t (just x)  = inj₂  x
fromMaybe t nothing   = inj₁  tt

maybe : {G : Ctx} →  Typ G (⋆ ⇒ ⋆)
maybe {G} = Lam (Data  
  (mkDT  (maybeDef (Var VZ))
         (\ e    → Maybe (interp (Var VZ) e))
         (\ {e}  → toMaybe    {⋆ ∷ G} {e} (Var VZ))
         (\ {e}  → fromMaybe  {⋆ ∷ G} {e} (Var VZ))))

vecDef : {G : Ctx} → Typ G ⋆ → (n : ℕ) → Typ G ⋆
vecDef _  0        = Con Unit
vecDef t  (suc n)  = 
  (App (App (Con Prod) t) (vecDef t n))

fromVec  :  {n : ℕ} {G : Ctx} 
            {t : Typ G ⋆} {e : Env G}
         → Vec (interp t e) n → (interp (vecDef t n) e)
fromVec {0}      []        = tt
fromVec {suc n}  (x ∷ xs)  = (x , fromVec xs)

toVec  : {n : ℕ} {G : Ctx} {t : Typ G ⋆} {e : Env G}
       → interp (vecDef t n) e → Vec (interp t e) n
toVec {0}     _         = []
toVec {suc n} (x , xs)  = (x ∷ toVec xs)

vec : {G : Ctx} → {n : ℕ} → Typ G (⋆ ⇒ ⋆)
vec {G} {n} = Lam (Data 
  (mkDT  (vecDef (Var VZ) n)
         (\ e  → Vec (interp (Var VZ) e) n)
         toVec
         fromVec))

{- The codes for recursive datatypes, like |List|, use this same
   idea. However, here we parameterize the structure with both the list
   parameter |ta| and the recursive instance of the list |tb|.
   Furthermore, the definitions of the code and the two coercions must be
   done mutually, as the type of the coercions depends on the code for
   lists.  Agda is not able to show that these definitions terminate
   because of this recursive structure---we are creating infinite objects
   to represent the recursive type.
 -}

{- This version can't be shown to terminate -}
{-
listDef  : {G : Ctx} → Typ G ⋆ 
         → (Typ G (⋆ ⇒ ⋆)) → Typ G ⋆
listDef ta tb = 
  (App (App (Con Sum) (Con Unit))
       (App (App (Con Prod) ta) (App tb ta)))

mutual
  list : {G : Ctx} → Typ G (⋆ ⇒ ⋆)
  list {G} = Lam (Data 
    (mkDT  (listDef (Var VZ) list)
           (\ e   → List (interp (Var VZ) e))
           (\{e}  → toList    {⋆ ∷ G}  {Var VZ}  {e})
           (\{e}  → fromList  {⋆ ∷ G}  {Var VZ}  {e})))

  fromList  : {G : Ctx} {t : Typ G ⋆} {e : Env G}
            → List (interp t e) → (interp (listDef t list) e)
  fromList []        = inj₁ tt
  fromList (x ∷ xs)  = inj₂ (x , xs)

  toList  : {G : Ctx} {t : Typ G ⋆} {e : Env G}
          → interp (listDef t list) e → List (interp t e)
  toList (inj₁ _)         = []
  toList (inj₂ (x , xs))  = (x ∷ xs)
-}

{- but this version does terminate.  Here the objects represented are
   not just lists, but lists paired with a proof that they have some
   particular length.  This lets us use the vector representation.
   Of course, we'll get generic functions not on lists but on
   list-proof pairs, but that's not so different. -}

fromList2 : {n : ℕ} {G : Ctx} {t : Typ G ⋆} {e : Env G}
          → Σ (List (interp t e)) (\ l → length l ≡ n) 
          → (interp (vecDef t n) e)
fromList2 {0}      ([] , refl)       = tt
fromList2 {0}      (x ∷ xs , () )
fromList2 {suc ._} ((x ∷ xs) , refl) = 
  ( x , fromList2 {length xs} (xs , refl))
fromList2 {suc _}  ([] , () )

toList2 : {n : ℕ} {G : Ctx} {t : Typ G ⋆} {e : Env G}
        → interp (vecDef t n) e 
        → Σ (List (interp t e)) (\ l → length l ≡ n)
toList2 {0}     tt       = [] , refl
toList2 {suc n} (x , xs) with toList2 {n} xs 
toList2 {suc ._} (x , xs) | (l , refl) = x ∷ l , refl

list2 : {G : Ctx} → {n : ℕ} → Typ G (⋆ ⇒ ⋆)
list2 {G} {n} = Lam (Data
  (mkDT (vecDef (Var VZ) n)
        (\ e → Σ (List (interp (Var VZ) e)) (\ l → length l ≡ n))
        (toList2 {n})
        (fromList2 {n})))


interp∗ : ∀ {G u n} → (Typ G u) → (Vec₁ (Env G) n) → Vec₁ ⟦ u ⟧ n
interp∗ t vec = repeat₁ (interp t) ⊛₁ vec

_⟨_⟩_ : {n : ℕ} → (Vec₁ Set (suc n) → Set1) → (k : Kind) → Vec₁ ⟦ k ⟧ (suc n) → Set1
b ⟨ ⋆ ⟩ v   = b v
b ⟨ k1 ⇒ k2 ⟩ v = ∀⇒ \(a : Vec₁ ⟦ k1 ⟧ _) → b ⟨ k1 ⟩ a → b ⟨ k2 ⟩ (v ⊛₁ a)

_⟨⟨_⟩⟩_ : {n : ℕ} → (Vec₁ Set (suc n) → Set1) → (k : Kind) → Vec₁ ⟦ k ⟧ (suc n) → Set1
b ⟨⟨ ⋆ ⟩⟩ v   = b v
b ⟨⟨ k1 ⇒ k2 ⟩⟩ v =  (a : Vec₁ ⟦ k1 ⟧ _) → (b ⟨ k1 ⟩ a → b ⟨⟨ k2 ⟩⟩ (v ⊛₁ a))

k-curry : (k : Kind)  → {n : ℕ} → {b : Vec₁ Set (suc n) → Set1} 
          → { Vs : Vec₁ ⟦ k ⟧ (suc n) } 
          → b ⟨⟨ k ⟩⟩ Vs  → b ⟨ k ⟩ Vs
k-curry ⋆ x = x
k-curry (k1 ⇒ k2){n}{b}{Vs} x = 
  \⇒ {suc n}{ ⟦ k1 ⟧ } 
     {\ (X : Vec₁ ⟦ k1 ⟧ (suc n)) → b ⟨ k1 ⟩ X → b ⟨ k2 ⟩ (Vs ⊛₁ X)} 
      (\ As y → k-curry k2 {n}{b}{Vs ⊛₁ As} (x As y))

TyConstEnv : {n : ℕ} (b : Vec₁ Set (suc n) → Set1) → Set1 
TyConstEnv {n} b = 
  {k : Kind} (c : Const k) → b ⟨ k ⟩ (repeat₁ (interp-c c))

data NGEnv {n : ℕ} (b : Vec₁ Set (suc n) → Set1)
  : Ctx → Set1 where
   NNil   : NGEnv b []
   NCons  : {k : Kind}  {G : Ctx}
     → (a : Vec₁ ⟦ k ⟧ (suc n))
     → b ⟨ k ⟩ a 
     → NGEnv b G 
     → NGEnv b (k ∷ G)

transpose : ∀ {n : ℕ}{G : Ctx}
   → {b : Vec₁ Set (suc n) → Set1} 
   → NGEnv b G → Vec₁ (Env G) (suc n)
transpose (NCons a _ ne) = 
  (repeat₁ _∷_) ⊛₁ a ⊛₁ (transpose ne)
transpose NNil = repeat₁ []

==-KIT : {n : ℕ} {b : Vec₁ Set (suc n) → Set1} 
      { k : Kind } {t1 t2 : Vec₁ ⟦ k ⟧ (suc n)} 
      → t1 ≡₁ t2 
      → b ⟨ k ⟩ t1 
      → b ⟨ k ⟩ t2
==-KIT refl x = x

c1 : {n : ℕ} { k : Kind } {G : Ctx}
    → (a : Vec₁ ⟦ k ⟧ n)
    → (envs : Vec₁ (Env G) n) 
    → a ≡₁ interp∗ (Var VZ) (repeat₁ _∷_ ⊛₁ a ⊛₁ envs)
c1 {zero}  []        [] = refl
c1 {suc n} (t ∷ ts)  ( x ∷ xs ) = ≡-tail₁ (c1 {n} ts xs)

c2 : { n : ℕ } { k : Kind }{k' : Kind} {G : Ctx}
      → (x : TyVar G k')
      → (t1 : Vec₁ ⟦ k ⟧ n) 
      → (envs : Vec₁ (Env G) n) 
      → interp∗ (Var x) envs ≡₁
        interp∗ (Var (VS x)) (repeat₁ _∷_ ⊛₁ t1 ⊛₁ envs)
c2 {zero}  x []       []       = refl
c2 {suc n} x (t ∷ ts) (y ∷ ys) = ≡-tail₁ (c2 x ts ys)

nLookup : {n : ℕ} {b : Vec₁ Set (suc n) → Set1}
          {k : Kind} {G : Ctx} 
          → (v : TyVar G k) 
          → (ce : NGEnv b G) 
          → b ⟨ k ⟩ (interp∗ (Var v) (transpose ce))
nLookup VZ  (NCons a e ce) = 
    ==-KIT (c1 a (transpose ce))     e
nLookup (VS x) (NCons a _ ce) = 
    ==-KIT (c2 x a (transpose ce))   (nLookup x ce)

c3 : {n : ℕ}{k k' : Kind}{G : Ctx}
    → (t : Typ (k' ∷ G) k)
    → (envs : Vec₁ (Env G) n) 
    → (as : Vec₁ ⟦ k' ⟧ n) 
    → (interp∗ t (repeat₁ _∷_ ⊛₁ as ⊛₁ envs)) ≡₁
      (interp∗ (Lam t) envs) ⊛₁ as
c3 {zero} t [] [] = refl 
c3 {suc n} t (a ∷ as) (b ∷ bs) = ≡-tail₁ (c3 t  as bs)

c4 : {n : ℕ}{k1 k2 : Kind} {G : Ctx}
     → (t1 : Typ G (k1 ⇒ k2))
     → (t2 : Typ G k1)
     → (envs : Vec₁ (Env G) n) 
     → ((interp∗ t1 envs) ⊛₁ (interp∗ t2 envs)) ≡₁
       interp∗ (App t1 t2) envs
c4 {zero}  _  _  []       = refl
c4 {suc n} t1 t2 (a ∷ as) = ≡-tail₁ (c4 t1 t2 as)

c5 : {n : ℕ}{k : Kind}{G : Ctx}
  → (c : Const k)
  → (envs : Vec₁ (Env G) n) 
  → repeat₁ (interp-c c) ≡₁ interp∗ (Con c) envs
c5 {zero}  _ []       = refl
c5 {suc n} c (a ∷ as) = ≡-tail₁ (c5 c as)

DataGen : {n : ℕ} → (b : Vec₁ Set (suc n) → Set1) → Set1
DataGen {n} b =
     {G : Ctx} 
  →  (dt : DT G) 
  →  (envs : Vec₁ (Env G) (suc n)) 
  →  b (repeat₁ (interp (DT-t dt)) ⊛₁ envs)
  →  b (repeat₁ (DT-s dt) ⊛₁ envs)

ngen-open  :  {n : ℕ} {b : Vec₁ Set (suc n) → Set1}
              {k : Kind} {G : Ctx}
           → (t : Typ G k) → (nge : NGEnv b G)
           → (ce : TyConstEnv b) → (d : DataGen b) 
           → b ⟨ k ⟩ (interp∗ t (transpose nge))
ngen-open (Var x)      nge ce d  = nLookup x nge
ngen-open{n}{b} (Lam {k1 = k1} t)      nge ce d  = 
  \⇒ \ (a : Vec₁ ⟦ k1 ⟧ (suc n)) 
  (nwt : b ⟨ k1 ⟩ a) → 
  ==-KIT (c3 t (transpose nge) a) 
  (ngen-open t (NCons a nwt nge) ce d)
ngen-open{n}{b}{k2}{G} (App{k1 = k1} t1 t2)  nge ce d  = 
  ==-KIT (c4 t1 t2 (transpose nge)) 
  (/⇒ (suc n) { ⟦ k1 ⟧ } { \ a → (b ⟨ k1 ⟩ a → 
     b ⟨ k2 ⟩ ((interp∗ t1 (transpose nge)) ⊛₁ a))}
  (ngen-open{n}{b}{k1 ⇒ k2}{G} t1 nge ce d) (interp∗ t2 (transpose nge)) 
  (ngen-open t2 nge ce d))
ngen-open (Con c) nge ce d      = 
  ==-KIT (c5 c (transpose nge)) (ce c) 
ngen-open (Data (mkDT t s to from)) nge ce d =
  d (mkDT t s to from) (transpose nge) (ngen-open t nge ce d)

c6 : { n : ℕ } { A B : Set1 } { f : A → B } { x : A } →
  (repeat₁{n} f) ⊛₁ repeat₁ x ≡₁ repeat₁ (f x)
c6 {zero} = refl
c6 {suc n} = ≡-tail₁ c6

ngen : {n : ℕ} {b : Vec₁ Set (suc n) → Set1}
       {k : Kind} (t : Typ [] k)
     → (TyConstEnv b) 
     → DataGen b 
     → b ⟨ k ⟩ (repeat₁ (interp t []))
ngen{n}{b}{k} t ce d = 
   ==-KIT {n}{b}{k} c6 (ngen-open t NNil ce d)

